from pathlib import Path

from unv.utils.tasks import register

from ...tasks import DeployComponentTasks
from ...settings import DeployComponentSettings

from ..systemd import SystemdTasksMixin


class RedisSettings(DeployComponentSettings):
    NAME = 'redis'
    SCHEMA = {
        'systemd': SystemdTasksMixin.SCHEMA,
        'config': {
            'type': 'dict',
            'schema': {
                'template': {'type': 'string', 'required': True},
                'name': {'type': 'string', 'required': True}
            }
        },
        'workdir': {'type': 'string', 'required': True},
        'port': {'type': 'integer', 'required': True},
        'maxmemory': {'type': 'string', 'required': True},
        'databases': {'type': 'integer', 'required': True},
        'root': {'type': 'string', 'required': True},
        'packages': {
            'type': 'dict',
            'schema': {
                'redis': {'type': 'string', 'required': True},
            },
            'required': True
        },
        'listen_private_ip': {'type': 'boolean', 'required': False},
        'iptables': {
            'type': 'dict',
            'schema': {
                'v4': {'type': 'string', 'required': True},
            },
            'required': True
        },
    }
    DEFAULT = {
        'systemd': {
            'template': 'server.service',
            'name': 'redis.service',
            'boot': True,
            'instances': {'count': 1}
        },
        'config': {
            # http://download.redis.io/redis-stable/redis.conf
            'template': 'server.conf',
            'name': 'redis.conf'
        },

        # TODO: move to base config (base package for this type of components)
        'workdir': '.',
        'port': 6379,
        'maxmemory': '128mb',
        'databases': 16,
        'root': 'app',
        'packages': {
            'redis': 'http://download.redis.io/releases/redis-5.0.5.tar.gz'
        },
        'iptables': {
            'v4': 'ipv4.rules'
        }
    }

    @property
    def build_dir(self):
        return self.home_abs / 'build'

    @property
    def bin(self):
        return self.root_abs / 'bin' / 'redis-server'

    @property
    def packages(self):
        return self._data['packages']

    @property
    def config_template(self):
        template = self._data['config']['template']
        if not template.startswith('/'):
            template = (self.local_root / template).resolve()
        return Path(template)

    @property
    def config_path(self):
        return self.root_abs / self._data['config']['name']

    @property
    def workdir(self):
        return self.root_abs / self._data['workdir']

    @property
    def port(self):
        return self._data['port']

    @property
    def maxmemory(self):
        return self._data['maxmemory']

    @property
    def databases(self):
        return self._data['databases']

    @property
    def iptables_v4_rules(self):
        return (self.local_root / self._data['iptables']['v4']).read_text()

    @property
    def listen_private_ip(self):
        return self._data.get('listen_private_ip', False)


SETTINGS = RedisSettings()


class RedisTasks(DeployComponentTasks, SystemdTasksMixin):
    SETTINGS = SETTINGS

    # TODO: add packages
    # TODO: add kernel systemd tasks to run settings on boot
    # # /proc/sys/net/core/somaxconn to 5000 (need command)
    async def get_iptables_template(self):
        return self.settings.iptables_v4_rules

    @register
    async def build(self):
        await self._create_user()

        async with self._cd(self.settings.build_dir, temporary=True):
            for package, url in self.settings.packages.items():
                await self._download_and_unpack(url, Path('.', package))

            async with self._cd('redis'):
                await self._run('make distclean')
                await self._run("make -j$(nproc) MALLOC=jemalloc")
                await self._run(
                    f"make PREFIX={self.settings.root_abs} install")

    @register
    async def sync(self):
        await self._upload_template(
            self.settings.config_template, self.settings.config_path)
        await self._sync_systemd_units()

    @register
    async def setup(self):
        await self.build()
        await self.sync()
        await self.start()



import time
import socket


from unv.utils.os import get_homepath
from unv.utils.tasks import register, Tasks

from ..settings import SETTINGS


def wait_ping(host, port, timeout=5):
    start = time.time()
    ready = False
    while time.time() - start < timeout:
        connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        connection.settimeout(1)
        result = connection.connect_ex((host, port))
        if result == 0:
            ready = True
            break

    return ready


class VagrantTasks(Tasks):
    NAMESPACE = 'vagrant'

    @register
    async def setup(self):
        # TODO: add config path to run commands from other dirs
        await self._local('vagrant destroy -f')
        await self._local('vagrant up')
        await self._local('vagrant status')
        await self.update_hosts()

    @register
    async def update_hosts(self):
        ips = [host['public_ip'] for _, host in SETTINGS.get_hosts()]
        known_hosts = get_homepath() / '.ssh' / 'known_hosts'

        for ip in ips:
            if not wait_ping(ip, 22):
                print(f"Can't setup ssh key for {ip} because host is down")
                print("After host up, please run vagrant.update_hosts")
                return

        if known_hosts.exists():
            with known_hosts.open('r+') as f:
                hosts = f.readlines()
                f.seek(0)
                for host in hosts:
                    if any(ip in host for ip in ips):
                        continue
                    f.write(host)
                f.truncate()

        for ip in ips:
            await self._local(f'ssh-keyscan {ip} >> {known_hosts}')



from pathlib import Path

from unv.deploy.tasks import (
    DeployComponentSettings, DeployComponentTasks, register
)


class PostgresSettings(DeployComponentSettings):
    NAME = 'postgres'
    SCHEMA = {
        'user': {'type': 'string', 'required': True},
        'work_dir': {'type': 'string', 'required': True},
        'build_dir': {'type': 'string', 'required': True},
        'sources': {
            'type': 'dict',
            'schema': {
                'postgres': {'type': 'string', 'required': True},
            },
            'required': True
        },
    }
    DEFAULT = {
        'user': 'postgres',
        'work_dir': '',
        'build_dir': 'build',
        'sources': {
            'postgres': 'https://ftp.postgresql.org/pub/source'
                        '/v12.0/postgresql-12.0.tar.gz',
        },
        'contrib': {
            # https://www.postgresql.org/docs/current/adminpack.html
            'adminpack': {'enabled': True},

            # https://www.postgresql.org/docs/current/amcheck.html
            'amcheck': {'enabled': True},

            # https://www.postgresql.org/docs/current/auth-delay.html
            'auth_delay': {'enabled': True},

            # https://www.postgresql.org/docs/current/auto-explain.html
            'auto_explain': {'enabled': True},

            # https://www.postgresql.org/docs/current/bloom.html
            'bloom': {'enabled': True},

            # btree_gin, https://www.postgresql.org/docs/current/btree-gin.html
            # btree_gist, citext, cube
        }
    }

    @property
    def sources(self):
        return self._data['sources']

    @property
    def work_dir(self):
        return self._data['work_dir']

    @property
    def build_dir(self):
        return self._data['build_dir']


class PostgresTasks(DeployComponentTasks):
    SETTINGS = PostgresSettings()

    @register
    async def build(self):
        await self._create_user()

        # await self._apt_install(
        #     'build-essential flex bison libreadline6-dev '
        #     'zlib1g-dev libossp-uuid-dev'
        # )
        # await self._mkdir(self.settings.build_dir, delete=True)
        async with self._cd(self.settings.build_dir):
            # for package, url in self.settings.sources.items():
            #     await self._download_and_unpack(url, Path('.', package))
            async with self._cd('postgres'):
                async with self._cd('contrib'):
                    await self._run('make')
                # for contrib in str(await self._run('ls contrib')).split():
                #     if contrib 
                #     async with self._cd(f'contrib/{contrib}'):
                #         await self._run('ls')
                await self._run('./configure --prefix=/home/postgres/app')
                await self._run('make')
                await self._run('make install')



import os
import sys
import importlib

from pathlib import Path


def run():
    sys.path.append(str(Path().cwd()))
    name, commands = sys.argv[1], sys.argv[2:]
    module_path = None
    modules = ['app.settings.', 'secure.', '']
    for module in modules:
        module_path = f'{module}{name}'
        try:
            importlib.import_module(module_path)
            break
        except (ImportError, ModuleNotFoundError):
            continue
    if not module_path:
        raise ValueError(f'Settings "{name}" not found in modules {modules}')
    os.environ['SETTINGS'] = module_path

    from .tasks import DeployTasksManager
    manager = DeployTasksManager()
    manager.register_from_settings()
    manager.run(' '.join(commands))



import copy
import inspect
import importlib

from pathlib import Path

from unv.utils.collections import update_dict_recur
from unv.app.settings import ComponentSettings, validate_schema


class DeploySettings(ComponentSettings):
    KEY = 'deploy'
    SCHEMA = {
        'tasks': {
            'type': 'list',
            'schema': {'type': 'string'}
        },
        'hosts': {
            'type': 'dict',
            'keyschema': {'type': 'string'},
            'valueschema': {
                'type': 'dict',
                'schema': {
                    'public_ip': {'type': 'string', 'required': True},
                    'private_ip': {'type': 'string'},
                    'port': {'type': 'integer'},
                    'provider': {'type': 'string'},
                    'tags': {'type': 'list', 'schema': {'type': 'string'}},
                    'settings': {'type': 'dict', 'required': False},
                    'components': {
                        'type': 'list',
                        'schema': {'type': 'string'}
                    }
                }
            }
        },
        'components': {'required': True, 'allow_unknown': True},
        'tags': {'type': 'dict'}
    }
    DEFAULT = {
        'tasks': [],
        'hosts': {},
        'components': {},
        'tags': {}
    }

    def get_hosts(self, component=''):
        for key, value in self._data['hosts'].items():
            if component in value.get('components', []) or not component:
                value.setdefault('private_ip', value['public_ip'])
                value.setdefault('port', 22)
                yield key, value

    def get_components(self, public_ip):
        for value in self._data['hosts'].values():
            if value['public_ip'] == public_ip:
                return value['components']
        return []

    def get_component_user(self, name):
        return self._data.get(name, {}).get('user', name)

    def get_component_settings(self, name):
        return self._data['components'].get(name, {})

    def get_tags_settings(self, name):
        return self._data['tags'].get(name, {})

    def get_host_override_settings(self, host) -> dict:
        for value in self._data['hosts'].values():
            if value['private_ip'] == host['private_ip'] and \
                    value['public_ip'] == host['public_ip']:

                settings = {}
                for tag in value.get('tags', []):
                    tag_settings = self.get_tags_settings(tag)
                    settings = update_dict_recur(settings, tag_settings)
                return update_dict_recur(settings, value.get('settings', {}))
        return {}

    @property
    def task_classes(self):
        for module_path in self._data['tasks']:
            module_path, class_path = module_path.split(':')
            yield getattr(importlib.import_module(module_path), class_path)


SETTINGS = DeploySettings()


class DeployComponentSettings:
    NAME = ''
    DEFAULT = {}
    SCHEMA = {}

    def __init__(self, settings=None):
        cls = self.__class__
        if settings is None:
            settings = SETTINGS.get_component_settings(cls.NAME)

        settings = update_dict_recur(cls.DEFAULT, settings, copy=True)
        settings = validate_schema(cls.SCHEMA, settings)

        self.local_root = Path(inspect.getfile(cls)).parent
        self._data = settings

    @property
    def user(self):
        return self._data.get('user', self.__class__.NAME)

    @property
    def enabled(self):
        if 'enabled' in self._data:
            return self._data['enabled']
        for _, host in SETTINGS.get_hosts():
            if self.__class__.NAME in host['components']:
                return True
        return False

    @property
    def home(self):
        return Path('~')

    @property
    def home_abs(self):
        return Path('/', 'home', self.user)

    @property
    def systemd(self):
        return self._data.get('systemd', {})

    @property
    def systemd_config(self):
        # TODO: move to class all systemd stuff
        return self.systemd.get('config', [])

    @property
    def systemd_dir(self):
        return Path('/etc', 'systemd', 'system')

    @property
    def systemd_type(self):
        return self.systemd.get('type', 'simple')

    @property
    def root(self):
        return self.home / self._data['root']

    @property
    def root_abs(self):
        return self.home_abs / self._data['root']

    def create_host_settings_copy(self, host):
        cls = self.__class__
        host_settings = SETTINGS.get_host_override_settings(host)
        host_settings = host_settings.get(cls.NAME, {})
        settings = update_dict_recur(self._data, host_settings, copy=True)
        return cls(settings)



import asyncio

from pathlib import Path

from watchgod import awatch

from ...tasks import DeployComponentTasks, nohost, register, onehost
from ...settings import SETTINGS, DeployComponentSettings

from ..python import PythonTasks, PythonSettings
from ..systemd import SystemdTasksMixin


class AppSettings(DeployComponentSettings):
    NAME = 'app'
    SCHEMA = {
        'bin': {'type': 'string'},
        'instance': {'type': 'integer'},
        'settings': {'type': 'string'},
        'systemd': SystemdTasksMixin.SCHEMA,
        'watch': {
            'type': 'dict',
            'schema': {
                'dirs': {'type': 'list', 'schema': {'type': 'string'}},
                'exclude': {'type': 'list', 'schema': {'type': 'string'}}
            }
        }
    }
    DEFAULT = {
        'bin': 'app',
        'instance': 1,
        'settings': 'secure.prod',
        'systemd': {
            'template': 'app.service',
            'name': 'app_{instance}.service',
            'boot': True,
            'type': 'simple',
            'instances': {'count': 0, 'percent': 0},
            'context': {
                'limit_nofile': 2000,
                'description': "Application description",
            }
        },
        'watch': {
            'dirs': ['./src', './secure'],
            'exclude': ['__pycache__', '*.egg-info']
        }
    }

    @property
    def python(self):
        settings = self._data.get('python', {})
        settings['user'] = self.user
        return PythonSettings(settings)

    @property
    def bin(self):
        return str(self.python.root_abs / 'bin' / self._data['bin'])

    @property
    def module(self):
        return self._data['settings']

    @property
    def instance(self):
        return self._data['instance']

    @property
    def watch_dirs(self):
        return (Path(directory) for directory in self._data['watch']['dirs'])

    @property
    def watch_exclude(self):
        return self._data['watch']['exclude']


class AppTasks(DeployComponentTasks, SystemdTasksMixin):
    SETTINGS = AppSettings()

    def __init__(self, manager, lock, user, host, settings=None):
        super().__init__(manager, lock, user, host, settings)
        self._python = PythonTasks(
            manager, lock, user, host, self.settings.python)

    @register
    @nohost
    async def watch(self):
        await asyncio.gather(*[
            self._watch_and_sync_dir(directory)
            for directory in self.settings.watch_dirs
        ])

    async def _watch_and_sync_dir(self, directory):
        site_packages_abs = self.settings.python.site_packages_abs
        async for _ in awatch(directory):
            for _, host in SETTINGS.get_hosts(self.NAMESPACE):
                with self._set_user(self.settings.user), \
                        self._set_host(host):
                    for sub_dir in directory.iterdir():
                        if not sub_dir.is_dir():
                            continue
                        await self._rsync(
                            sub_dir, site_packages_abs / sub_dir.name,
                            self.settings.watch_exclude
                        )
                    await self.restart()

    @register
    async def build(self):
        await self._create_user()
        await self._python.build()

    @register
    @onehost
    async def shell(self):
        return await self._python.shell(
            prefix=f'SETTINGS={self.settings.module}')

    @register
    async def sync(self, type_=''):
        flag = '-I' if type_ == 'force' else '-U'
        name = (await self._local('python setup.py --name')).strip()
        version = (await self._local('python setup.py --version')).strip()
        package = f'{name}-{version}.tar.gz'

        # once for all deploy
        async with self._lock:
            await self._local('pip install -e .')
            await self._local('python setup.py sdist bdist_wheel')
            await self._upload(Path('dist', package))
            await self._local('rm -rf ./build ./dist')

        await self._python.pip(f'install {flag} {package}')
        await self._rmrf(Path(package))

        # generate app settings from deploy settings with create command
        # pop deploy settings
        await self._rsync(Path('secure'), Path('secure'))

        await self._sync_systemd_units()

    @register
    async def setup(self):
        await self.build()
        await self.sync()
        await self.start()



import asyncio
import logging
import functools

from contextlib import contextmanager, asynccontextmanager
from pathlib import Path

import jinja2

from unv.utils.tasks import Tasks, TasksManager, TaskRunError, register

from .settings import SETTINGS, DeployComponentSettings


def as_root(func):
    """Task will run from root, sets to self.user."""

    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        with self._set_user('root'):
            return await func(self, *args, **kwargs)

    return wrapper


def nohost(task):
    task.__nohost__ = True
    return task


def onehost(task):
    task.__onehost__ = True
    return task


class DeployTasks(Tasks):
    def __init__(self, manager, lock, user, host):
        self.user = user
        self.public_ip = host['public_ip']
        self.private_ip = host['private_ip']
        self.port = host['port']

        self._current_prefix = ''
        self._logger = logging.getLogger(self.__class__.__name__)
        self._lock = lock

        super().__init__(manager)

    async def _calc_instances_count(self, count: int = 0, percent: int = 0):
        if percent:
            cpu_cores = int(await self._run('nproc --all'))
            cpu_cores = int(cpu_cores / 100.0 * percent)
            cpu_cores += count
            count = cpu_cores
        return count or 1

    def get_all_deploy_tasks(self):
        for task_class in self._manager.tasks.values():
            if issubclass(task_class, DeployTasks):
                yield task_class(self._manager, self._lock, self.user, {
                    'public_ip': self.public_ip,
                    'private_ip': self.private_ip,
                    'port': self.port
                })

    @contextmanager
    def _set_user(self, user):
        old_user = self.user
        self.user = user
        try:
            yield self
        finally:
            self.user = old_user

    @contextmanager
    def _set_host(self, host):
        old_public_ip, old_private_ip, old_port =\
            self.public_ip, self.private_ip, self.port
        self.public_ip, self.private_ip, self.port =\
            host['public_ip'], host['private_ip'], host['port']
        try:
            yield self
        finally:
            self.public_ip, self.private_ip, self.port =\
                old_public_ip, old_private_ip, old_port

    @contextmanager
    def _prefix(self, command):
        old_prefix = self._current_prefix
        self._current_prefix = f'{self._current_prefix} {command} '
        try:
            yield
        finally:
            self._current_prefix = old_prefix

    @asynccontextmanager
    async def _cd(self, path: Path, temporary=False):
        if temporary:
            await self._mkdir(path, delete=True)
        try:
            with self._prefix(f'cd {path} &&'):
                yield
        finally:
            if temporary:
                await self._rmrf(path)

    @as_root
    async def _sudo(self, command, strip=True):
        """Run command on server as root user."""
        return await self._run(command, strip)

    async def _wait(self, timeout: int):
        await asyncio.sleep(timeout)

    async def _reboot(self, timeout: int = 20):
        try:
            await self._sudo('reboot')
        except TaskRunError:
            pass
        await self._wait(timeout)

    async def _create_user(self):
        """Create user if not exist and sync ssh keys."""
        user = self.user

        with self._set_user('root'):
            try:
                await self._run("id -u {}".format(user))
            except TaskRunError:
                await self._run(
                    "adduser --quiet --disabled-password"
                    " --gecos \"{0}\" {0}".format(user)
                )

                local_ssh_public_key = Path('~/.ssh/id_rsa.pub')
                local_ssh_public_key = local_ssh_public_key.expanduser()
                keys_path = Path(
                    '/', 'home' if user != 'root' else '',
                    user, '.ssh'
                )

                await self._mkdir(keys_path)
                await self._run(f'chown -hR {user} {keys_path}')
                await self._run('echo "{}" >> {}'.format(
                    local_ssh_public_key.read_text().strip(),
                    keys_path / 'authorized_keys'
                ))

    @as_root
    async def _apt_install(self, *packages):
        with self._prefix('DEBIAN_FRONTEND=noninteractive'):
            await self._run('apt-get update -y -q')
            await self._run(
                'apt-get install -y -q --no-install-recommends '
                '--no-install-suggests {}'.format(' '.join(packages))
            )

    async def _run(self, command, strip=True, interactive=False) -> str:
        self._logger.debug(
            f'run [{self.user}@{self.public_ip}:{self.port}] '
            f'{self._current_prefix}{command}'
        )
        interactive_flag = '-t' if interactive else ''
        response = await self._local(
            f"ssh {interactive_flag} -p {self.port} "
            f"{self.user}@{self.public_ip} "
            f"'{self._current_prefix}{command}'",
            interactive=interactive
        ) or ''
        if strip:
            response = response.strip()
        return response

    async def _rmrf(self, path: Path):
        await self._run(f'rm -rf {path}')

    async def _mkdir(self, path: Path, delete=False):
        if delete:
            await self._rmrf(path)
        await self._run(f'mkdir -p {path}')

    async def _upload(self, local_path: Path, path: Path = '~/'):
        await self._local(
            f'scp -r -P {self.port} {local_path} '
            f'{self.user}@{self.public_ip}:{path}'
        )

    async def _rsync(self, local_dir, root_dir, exclude=None):
        exclude = [f"--exclude '{path}'" for path in exclude or []]
        exclude = ' '.join(exclude)
        await self._local(
            f'rsync -rave "ssh -p {self.port}" --delete {exclude} '
            f'{local_dir}/ {self.user}@{self.public_ip}:{root_dir}'
        )

    async def _upload_template(
            self, local_path: Path, path: Path, context: dict = None):
        context = context or {}
        context.setdefault('deploy', self)

        async with self._lock:
            render_path = Path(f'{local_path}.render')
            template = jinja2.Template(
                local_path.read_text(), enable_async=True)
            render_path.write_text(await template.render_async(context))
            try:
                await self._upload(render_path, path)
            finally:
                render_path.unlink()

    async def _download_and_unpack(
            self, url: str, dest_dir: Path = Path('.'),
            archive_dir_name: str = ''):
        await self._run(f'wget -q {url}')
        archive = url.split('/')[-1]
        await self._run(f'tar xf {archive}')
        archive_dir = archive_dir_name or archive.split('.tar')[0]

        if '*' in archive_dir:
            new_archive_dir = await self._run(
                f"ls -d */ | grep '{archive_dir}'")
            if not new_archive_dir:
                raise ValueError(f'No archive dir found {new_archive_dir}')
            archive_dir = new_archive_dir.rstrip('/')

        await self._mkdir(dest_dir)
        await self._run(f'mv {archive_dir}/* {dest_dir}')

        await self._rmrf(archive)
        await self._rmrf(archive_dir)

    @register
    @onehost
    async def ssh(self):
        return await self._run('bash', interactive=True)

    @register
    @onehost
    @as_root
    async def root(self):
        return await self._run('bash', interactive=True)


class DeployComponentTasks(DeployTasks):
    SETTINGS = None

    def __init__(self, manager, lock, user, host, settings=None):
        settings = settings or self.__class__.SETTINGS
        if settings is None or \
                not isinstance(settings, DeployComponentSettings):
            raise ValueError(
                "Provide correct 'SETTINGS' value "
                "should be an instance of class 'DeployComponentSettings' not "
                f"[{settings}] value and type {type(settings)}"
            )
        self.settings = settings.create_host_settings_copy(host)

        super().__init__(manager, lock, self.settings.user, host)

    @classmethod
    def get_namespace(cls):
        return cls.SETTINGS.NAME


class DeployTasksManager(TasksManager):
    def register_from_settings(self):
        for class_ in SETTINGS.task_classes:
            self.register(class_)

    def run_task(self, task_class, name, args):
        method = getattr(task_class, name)
        is_nohost = getattr(method, '__nohost__', False)
        is_onehost = getattr(method, '__onehost__', False)

        if issubclass(task_class, DeployTasks):
            if is_nohost:
                user = '__nohost__'
                hosts = [
                    ('', {'public_ip': None, 'private_ip': None, 'port': 0})
                ]
            else:
                user, hosts = self._get_user_with_hosts(task_class)

            if is_onehost and len(hosts) > 1:
                hosts_per_index = []
                for index, (host_name, host) in enumerate(hosts, start=1):
                    hosts_per_index.append([host_name, host])
                    print(f" ({index}) - {host_name} [{host['public_ip']}]")
                chosen_index = int(input('Please choose host to run task: '))
                hosts = [hosts_per_index[chosen_index + 1]]

            async def run():
                lock = asyncio.Lock()
                tasks = [
                    getattr(task_class(self, lock, user, host), name)
                    for _, host in hosts
                ]
                await asyncio.gather(*[task(*args) for task in tasks])

            asyncio.run(run())
        else:
            return super().run_task(task_class, name, args)

    def _get_user_with_hosts(self, task_class):
        name = task_class.get_namespace()
        return (
            SETTINGS.get_component_user(name),
            list(SETTINGS.get_hosts(name))
        )



from pathlib import Path

from unv.utils.tasks import register

from ...tasks import DeployComponentTasks, as_root
from ...settings import DeployComponentSettings

from ..systemd import SystemdTasksMixin


class NginxSettings(DeployComponentSettings):
    NAME = 'nginx'
    SCHEMA = {
        'systemd': SystemdTasksMixin.SCHEMA,
        'master': {'type': 'boolean', 'required': True},
        'root': {'type': 'string', 'required': True},
        'packages': {
            'type': 'dict',
            'schema': {
                'nginx': {'type': 'string', 'required': True},
                'pcre': {'type': 'string', 'required': True},
                'zlib': {'type': 'string', 'required': True},
                'openssl': {'type': 'string', 'required': True},
                'geoip2': {'type': 'string', 'required': True},
                'libmaxminddb': {'type': 'string', 'required': True}
            },
            'required': True
        },
        'packages_dir': {
            'type': 'dict',
            'required': False
        },
        'geoip2db': {
            'type': 'dict',
            'schema': {
                'city': {'type': 'string', 'required': True},
                'country': {'type': 'string', 'required': True},
                'lang': {
                    'type': 'string',
                    'required': True,
                    'allowed': ['en', 'ru']
                }
            },
            'required': True
        },
        'configs': {'type': 'dict'},
        'connections': {'type': 'integer', 'required': True},
        'workers': {'type': 'integer', 'required': True},
        'aio': {'type': 'boolean', 'required': True},
        'sendfile': {'type': 'boolean', 'required': True},
        'tcp_nopush': {'type': 'boolean', 'required': True},
        'tcp_nodelay': {'type': 'boolean', 'required': True},
        'keepalive_timeout': {'type': 'integer', 'required': True},
        'include': {'type': 'string', 'required': True},
        'access_log': {'type': 'string', 'required': True},
        'error_log': {'type': 'string', 'required': True},
        'default_type': {'type': 'string', 'required': True},
        'geoip2': {'type': 'boolean', 'required': True},
        'iptables': {
            'type': 'dict',
            'schema': {
                'v4': {'type': 'string', 'required': True}
            },
            'required': True
        }
    }
    DEFAULT = {
        'systemd': {
            'template': 'server.service',
            'name': 'nginx.service',
            'boot': True,
            'instances': {'count': 1}
        },
        'master': True,
        'root': 'app',
        'packages': {
            'nginx': 'http://nginx.org/download/nginx-1.17.1.tar.gz',
            'pcre': 'https://ftp.pcre.org/pub/pcre/pcre-8.42.tar.gz',
            'zlib': 'http://www.zlib.net/zlib-1.2.11.tar.gz',
            'openssl': 'https://www.openssl.org/source/openssl-1.1.1a.tar.gz',
            'geoip2': 'https://github.com/leev/ngx_http_geoip2_'
                'module/archive/master.tar.gz',
            'libmaxminddb': 'https://github.com/maxmind/libmaxminddb/releases'
                '/download/1.3.2/libmaxminddb-1.3.2.tar.gz',
        },
        'packages_dir': {
            'geoip2': 'ngx_http_geoip2_module-master',
        },
        'geoip2db': {
            'city': 'https://geolite.maxmind.com/download/geoip/database/'
                'GeoLite2-City.tar.gz',
            'country': 'https://geolite.maxmind.com/download/geoip/database/'
                'GeoLite2-Country.tar.gz',
            'lang': 'en'
        },
        'configs': {'server.conf': 'nginx.conf'},
        'connections': 1000,
        'workers': 1,
        'aio': True,
        'sendfile': True,
        'tcp_nopush': True,
        'tcp_nodelay': True,
        'keepalive_timeout': 60,
        'include': 'conf/apps/*.conf',
        'access_log': 'logs/access.log',
        'error_log': 'logs/error.log',
        'default_type': 'application/octet-stream',
        'geoip2': False,
        'iptables': {'v4': 'ipv4.rules'}
    }

    @property
    def build(self):
        return self.root / 'build'

    @property
    def packages(self):
        return self._data['packages']

    @property
    def packages_dir(self):
        return self._data.get('packages_dir', {})

    @property
    def configs(self):
        for template, name in self._data['configs'].items():
            if not template.startswith('/'):
                template = (self.local_root / template).resolve()
            yield Path(template), self.root / 'conf' / name

    @property
    def include(self):
        return self.root_abs / self._data['include']

    @property
    def access_log(self):
        return self.root_abs / self._data['access_log']

    @property
    def error_log(self):
        return self.root_abs / self._data['error_log']

    @property
    def default_type(self):
        return self._data['default_type']

    @property
    def aio(self):
        return 'on' if self._data['aio'] else 'off'

    @property
    def sendfile(self):
        return self._data['sendfile']

    @property
    def tcp_nopush(self):
        return self._data['tcp_nopush']

    @property
    def tcp_nodelay(self):
        return self._data['tcp_nodelay']

    @property
    def keepalive_timeout(self):
        return self._data['keepalive_timeout']

    @property
    def workers(self):
        return self._data['workers']

    @property
    def connections(self):
        return self._data['connections']

    @property
    def master(self):
        return self._data['master']

    @property
    def geoip2(self):
        return self._data['geoip2']

    @property
    def geoip2_city_path(self):
        return self.root_abs / 'geoip2' / 'GeoLite2-City.mmdb'

    @property
    def geoip2_country_path(self):
        return self.root_abs / 'geoip2' / 'GeoLite2-Country.mmdb'

    @property
    def geoip2_lang(self):
        return self._data['geoip2db']['lang']

    @property
    def geoip2db_city_url(self):
        return self._data['geoip2db']['city']

    @property
    def geoip2db_country_url(self):
        return self._data['geoip2db']['country']

    @property
    def iptables_v4_rules(self):
        return (self.local_root / self._data['iptables']['v4']).read_text()


class NginxTasks(DeployComponentTasks, SystemdTasksMixin):
    SETTINGS = NginxSettings()

    async def get_iptables_template(self):
        return self.settings.iptables_v4_rules

    @as_root
    async def _install_libmaxminddb(self):
        # TODO: move to other package?
        package = 'libmaxminddb'
        url = self.settings.packages[package]

        async with self._cd('build', temporary=True):
            await self._download_and_unpack(url, Path('.', package))

            async with self._cd('libmaxminddb'):
                await self._run('./configure')
                await self._run('make')
                await self._run('make check')
                await self._run('make install')
                await self._run('ldconfig')

    @register
    async def build(self):
        if not self.settings.master:
            print('Nginx already builded on this host, just use nginx.sync')
            return

        await self._create_user()
        await self._mkdir(self.settings.include.parent)
        await self._apt_install(
            'build-essential', 'autotools-dev', 'libexpat-dev',
            'libgd-dev', 'libgeoip-dev', 'liblua5.1-0-dev',
            'libmhash-dev', 'libpam0g-dev', 'libperl-dev',
            'libxslt1-dev'
        )

        if self.settings.geoip2:
            await self._install_libmaxminddb()

            async with self._cd(self.settings.root):
                await self._mkdir('geoip2')
                async with self._cd('geoip2'):
                    await self._download_and_unpack(
                        self.settings.geoip2db_city_url,
                        archive_dir_name='GeoLite2-City_*'
                    )
                    await self._download_and_unpack(
                        self.settings.geoip2db_country_url,
                        archive_dir_name='GeoLite2-Country_*'
                    )
                    await self._run('rm *.txt')

        async with self._cd(self.settings.build, temporary=True):
            for package, url in self.settings.packages.items():
                if package == 'libmaxminddb':
                    continue
                if package == 'geoip2' and not self.settings.geoip2:
                    continue

                await self._download_and_unpack(
                    url, Path('.', package),
                    archive_dir_name=self.settings.packages_dir.get(package)
                )

            async with self._cd('nginx'):
                build_command = (
                    f"./configure --prefix={self.settings.root_abs} "
                    f"--user='{self.user}' --group='{self.user}' "
                    "--with-pcre=../pcre "
                    "--with-pcre-jit --with-zlib=../zlib "
                    "--with-openssl=../openssl --with-http_ssl_module "
                    "--with-http_v2_module --with-threads "
                    "--with-file-aio --with-http_realip_module "
                )
                if self.settings.geoip2:
                    build_command += '--add-module=../geoip2 '
                await self._run(build_command)
                await self._run('make')
                await self._run('make install')

    @register
    async def sync(self):
        for template, path in self.settings.configs:
            await self._upload_template(template, path)

        for task in self.get_all_deploy_tasks():
            get_configs = getattr(task, 'get_nginx_include_configs', None)
            if get_configs is not None:
                configs = await get_configs()
                for template, path in configs:
                    await self._upload_template(
                        template,
                        self.settings.root / self.settings.include.parent
                        / path, {'deploy': task, 'nginx_deploy': self}
                    )

        await self._sync_systemd_units()

    @register
    async def setup(self):
        await self.build()
        await self.sync()
        await self.start()



from unv.utils.tasks import register


class SystemdTasksMixin:
    SCHEMA = {
        'type': 'dict',
        'schema': {
            'template': {'type': 'string'},
            'name': {'type': 'string'},
            'boot': {'type': 'boolean'},
            'type': {'type': 'string'},
            'config': {
                'type': 'list',
                'schema': {
                    'type': 'string'
                }
            },
            'instances': {
                'type': 'dict',
                'schema': {
                    'count': {'type': 'integer'},
                    'percent': {'type': 'integer'}
                }
            },
            'context': {
                'type': 'dict',
                'schema': {
                    'limit_nofile': {'type': 'integer'},
                    'description': {'type': 'string'}
                }
            }
        }
    }

    async def _get_systemd_services(self):
        systemd = self.settings.systemd
        name = systemd['name']
        count = await self._calc_instances_count(**systemd['instances'])
        for instance in range(1, count + 1):
            service = systemd.copy()
            service['name'] = name.format(instance=instance)
            service['instance'] = instance
            yield service

    async def _sync_systemd_units(self):
        services = [service async for service in self._get_systemd_services()]

        for service in services:
            service_path = self.settings.systemd_dir / service['name']

            context = {'instance': service['instance']}.copy()
            context.update(service.get('context', {}))
            path = service['template']
            if not str(path).startswith('/'):
                path = (self.settings.local_root / service['template'])
                path = path.resolve()

            with self._set_user('root'):
                await self._upload_template(path, service_path, context)

        with self._set_user('root'):
            await self._run('systemctl daemon-reload')
            await self._systemctl('enable', boot_only=True)

    async def _systemctl(
            self, command: str, display=False, boot_only=False):
        results = []
        async for service in self._get_systemd_services():
            if 'manage' in service and not service['manage']:
                continue
            if boot_only and not service.get('boot', False):
                continue

            with self._set_user('root'):
                result = await self._run(
                    f'systemctl {command} {service["name"]}')
            results.append(result)
        return results

    @register
    async def start(self):
        await self._systemctl('start')

    @register
    async def stop(self):
        await self._systemctl('stop')

    @register
    async def restart(self):
        await self._systemctl('restart')

    @register
    async def status(self):
        results = await self._systemctl('status')
        for result in results:
            print(result)
